#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 19 15:19:29 2020
file split on Sun Oct 18 16:07:53 2020
@author: davidworrall
version 0.01: Fri Jun 19 15:19:29 2020. Set up intitial cv read
version 0.02: 2020 31 Define class structure for a country
version 0.03: 2020 08 06 Add SMA
version 0.04: 2020 09 18 finish marked-line plotting routines * non-printing vals <- 0
version 0.05: 2020 09 19 finish x-axis 7day divisions and labels
version 0.06: 2020 09 23 regional sort according to classification in WorldRegionsUpdated.csv
            complete plot generation and write to file
version 0.07 begin investigation of auto anomoly detection. found TAD (twitter ...)
see https://blog.twitter.com/engineering/en_us/a/2015/introducing-practical-and-robust-anomaly-detection-in-a-time-series.html

version 0.08 code Split code into 3 files:
    - covid19data.py  .... included here
    - covid19class.py .... (this file)
    -covid19plotting.py .. includes all of the above

databases in play:
  - deathsDB = deathsDB[1:] # drop the column headings
  - config.datesList = datePad + columnLabels[4:]
  - countriesList = []      # collates all multiple-province country provinces into a single country entry
  - config.provinceDeathsDict = {} # holds all country, province and summed-province data
  - config.populationDataList = []       # demographic data CONVER TO DICT

  - inJustDeathDB = []      # countries/provinces that are only in deaths_global.csv
  - inJustpopulationDataList = [] # countries/provinces that are only in countryPopulationData.csv
  - inBothDB = []
  - demographicDict = {}    # an amalgulm of the two .csv files data # regionName,countryName,countrySymbol,area,pop etc

""" 
import config                                                                   # our global variables initialization module

if config.reportProgress:                                                       # True or False
    from datetime import datetime
    productionTime = datetime.now()
    print ("=== Begin covid19cslass.py:\t", productionTime)

colLabelRow     = 0

regionNameCol   = 0
countryNameCol  = 1
countrySymbolCol= 2
latitudeCol     = 3
longitudeCol    = 4
areaCol         = 5
populationCol   = 6
densityCol      = 7                                                             #already calc'd from 5 &  
geoRegionCol    = 8
censusDateCol   = 9

import numpy as np

# ====================DEFINE DEATH CLASS FOR GLOBAL COUNTRIES/STATES?PROVINCES#

class CountryDeaths():    # define a class for handling Covid-19 death data for a country
    def __init__(self, countryName):
        #Initialize data attributes.
        self.countryName        = countryName
        self.countryExists      = False
        self.countryCode        = ''
        self.printData          = False                                         # used to filter death data when using print (instance)
        self.provinceName       = ''                                            # filled from deahDB  by __init__()
        self.region             = ''                                            # ahould be filled from demographicDataDict - Check!
        self.area               = 0
        self.population         = 0
        self.populationDensity  = 0
#        self.latitude           = countriesList[latCol_deathsData]
#        self.longitude          = countriesList[latCol_deathsData]
        if self.countryName in config.demographicDict.keys():                                    # defined in demographicData.py
            self.countryCode     = config.demographicDict [self.countryName][countrySymbolCol]   # Unique two letter code
            self.area            = config.demographicDict [self.countryName][areaCol]            # land area (in sq kms)
            print (self. countryName, 'Area:' , self.area)
            self.population      = int(config.demographicDict [self.countryName][populationCol]) # the population
            if self.population > 0:
                self.populationDensity = config.demographicDict [self.countryName][densityCol]   # pre-calculted in the .csv. See next line
#               self.populationDensity  = self.population/self.area                          # substitute this line if the above is not used
 #                                   #sun     mon       tues     
        self.prePadDates        = ['1/19/20','1/20/20', '1/21/20']              # 1/22/20 = first data date. prepad for a Sunday start
        self.prePadData         = [0] * len (self.prePadDates)                  # insert len ( self.prePadDates days of data = 0)
        self.cumulativeDeaths   = self.prePadData                               # start the list by prepadding
        self.deathIncrements    = []                                            # NB calculated from self.cumulativeDeaths, so no need to prePad
        self.ascending          = []
        self.sortedIndices      = []                                            # indices of increments in ascending order
        self.NrDataPoints       = 0                                             # Needed many times so no need to keep calculating it
        self.mean               = 0                                             # Mean nr of Deaths / day calculated in the Stats method
        self.median             = 0                                             # middle nr of Deaths / day calculated in the Stats method
        self.max                = 0                                             # Max Nr of Deaths in any one day calculated in the Stats method
        self.SMAshortPeriod     = 7                                             # calculated in the Stats method
        self.SMAlongPeriod      = 14                                            # calculated in the Stats method
        self.SMAshortArray      = np.array([])                                  # Simple Moving Average over self.SMAperioShort  
        self.SMAlongArray       = np.array([])                                  # Simple Moving Average over self.SMAperiods  
        self.anomolies2adjust   = 0                                             # non-zero if adjusted. Calc'ed for Regions in config.outlierProvinceList
        self.maxVals            = ()                                            # indices of the 3 largest daily death values # fix the values
        self.barChartColor      = '#FFFFFF'                                     # black, by default [still hot-linked but used to reveal if there is a color-mapping problem                                         
        self.plotColor          = '#FFFFFF'                                     # black, by default                                         

        if self.countryName not in config.provinceDeathsDict.keys():
            self.countryExists = False
            print (self.countryName + " does not exist in the Databases.")
            return (None)
        else:
            self.countryExists = True
#            print("*********", config.provinceDeathsDict[self.countryName])
#            print
            self.provinceName   = config.provinceDeathsDict[countryName][0]     # countriesList[self.countryName][0]
            self.latitude       = config.provinceDeathsDict[countryName][2]     # countriesList[latCol_deathsData] WRONG! this is empty! use demographic dict
            self.longitude      = config.provinceDeathsDict[countryName][3]     # countriesList[longCol_deathsData] WRONG! this is empty!  use demographic dict
#            self.area           = config.demographicDict [self.countryName][areaCol]
#            self.population     = config.demographicDict [self.countryName][populationCol]
#            if self.population > 0:
#                self.populationDensity  = self.population/self.area

            self.cumulativeDeaths       = self.prePadData + config.provinceDeathsDict[countryName][4:]
            # block the printing of values <=0
            for indx in range (len(self.cumulativeDeaths)):
                if self.cumulativeDeaths[indx] <= 0:
                    self.cumulativeDeaths[indx] = 0 # float ('nan')             # nan = not-a-number
            self.deathIncrements = [0]* len (self.cumulativeDeaths)
            for dayNr in range (1,len (self.cumulativeDeaths)):
                self.deathIncrements[dayNr] = self.cumulativeDeaths[dayNr]-self.cumulativeDeaths[dayNr-1]
                if self.deathIncrements[dayNr] <0:                              # indicates data error
                    if self.printData:
                        print ("DATA ERROR:",self.countryName , end="\t")
#                      print ("DATA ERROR:Some dead rose again in ",self.countryName , "on", :", end="/t")
                        print (dayNr,":",config.datesList[dayNr], "\t", self.cumulativeDeaths[dayNr-1],"\t",self.cumulativeDeaths[dayNr],"\tdiff :", self.deathIncrements[dayNr-1],self.deathIncrements[dayNr])

                    self.deathIncrements[dayNr] = 0  # just zero it : #  ============ ERROR CORRECTION==============
 #           self.generateStats()
            #  ============== NBNBNBNBNBNBN
            #print (self.deathIncrements)
            if self.provinceName == "":
                self.provinceName = self.countryName                            # the 'province' is the country
            self.NrDataPoints = len (self.deathIncrements)
            self.generateStats()                                                # execute before convertsion of 0's to NANS
            self.fixAnomalies()
            self.makeSMAarrays()                                                # make the short and long SMA arrays
#            self.zeros2NANs()                                                  # disable printing of 0's
   
#            countriesInConsiderationList += [self.countryName]

    def __str__(self):
#        # Defines what to do when the default python help() function is called on an instance
        if self.countryExists == True:
            print ('#' * 30)
            print ('Country:',                      self.countryName, end = ' ')
            print ('(Code: ',                       self.countryCode, ')')
#            print ('#' * 30)
            if self.provinceName !='' or (self.province == self.countryName) :
                print ('Province:\t' ,              self.provinceName)
            print ("Total Number of Deaths:\t",     self.cumulativeDeaths[-1])
            print ('µ (mean) of new deaths/day:\t', self.mean)
            print ('Max Nr of deaths/day:\t',       self.max )
            if self.countryName in config.demographicDict.keys():
                print ('Latitude:\t' ,              self.latitude)
                print ('Longitude:\t' ,             self.longitude)
 #               self.area                       = config.demographicDict [self.countryName][areaCol]
 #               self.population                 = config.demographicDict [self.countryName][populationCol]
                print ("CountryCode:\t",            self.countryCode)
                print ('Area (sq kms):\t',          self.area)
                print ('Population:\t' ,            self.population)
                if self.population > 0:
                    self.populationDensity  =       self.population/self.area
                    print ('Population Density:\t', self.populationDensity)
                else: print ("No demographic data available")
            if not self.printData:
                print ("Switch self.printData on for more complete details of "+self.countryName+ 'statistics.')
            else:                         # print the Dtat
                print (self.countryName+': Cumulative Deaths:\n',       self.cumulativeDeaths)
                print (self.countryName+': Death Increments:\n',        self.deathIncrements)
#                print (self.countryName+': Sorted deathIncrements:\n',  self.ascending)
        return ("")

    def generateStats (self):
        # generates some stats of the deaths. Perhaps run it twice - before and after fixing anomalies
        from statistics import mean, median
        self.median = int( median (self.deathIncrements))                       # not affected by leading zeros
        if any(self.deathIncrements):                                           # calc. only from the first non-zero day   
            indxof1stNonZero = next((i for i, x in enumerate(self.deathIncrements) if x), None) # find the index
            tmpIncrements = self.deathIncrements[indxof1stNonZero:]
            self.mean = int(round (mean(tmpIncrements)))            
#        self.mean = int(round (mean (self.deathIncrements)))
#        self.mean       = self.nanMean (self.deathIncrements) #; print ('Daily deaths, mean   :', self.mean)
        self.max        = max (self.deathIncrements) #; print ('Daily deaths, maxVal :', self.max )
        self.ascending  = sorted(self.deathIncrements)
#        sorted(range(len(vals)), key=vals.__getitem__)
        kee = self.deathIncrements.__getitem__
        deathIncrementIndices = range(len(self.deathIncrements))                # 0,1, 2, ... len()
        self.sortedIndices = sorted(deathIncrementIndices, key=kee)
#       alternative: sorted(range(len(vals)), key=vals.__getitem__)

    def valueIndexSort(self, listy):
        # ascending sort of listy. returns key value pairs
        sortedListy = sorted (listy)                                            # sort the items
        kee = listy.__getitem__                                                 # get the kee position 
        indices = range(len(self.deathIncrements))                              # 0,1, 2, ... len() 
        self.sortedIndices = sorted(deathIncrementIndices, key=kee)             # the sorted indices
        
    def fixAnomalies(self):
         if self.countryName in config.outlierProvinceList: 
             print ("Data Anomaly:", self.countryName, end = "\t")
             print ("[indexes, values]:", self.sortedIndices[-3:], self.ascending [-3:] )
             self.maxVals = self.sortedIndices[-3:]
             for indx in range(len(self.maxVals)): 
                 self.maxVals [indx] = (self.maxVals [indx],self.deathIncrements[self.maxVals [indx]] )
#            print (self.maxVals)   #indices of the 3 largest daily death values # fix the values
    
    def SMA (self, NrPeriods):
        #Computes a Simple Moving Average of self.deathIncrements over NrPeriods
        self.SMAperiod = NrPeriods
        self.SMAdata = []
        firstDataPointIndex = len (self.deathIncrements)                        # find first non-zero  datapoint
        lastDataPointIndex = len (self.deathIncrements)
        if lastDataPointIndex <  self.SMAperiod:
            print ("Averaging period too long." +
                   "Longest period is: ", lastDataPointIndex-self.SMAperiod)
            return(None)
        index = 0
        while index < self.NrDataPoints:
            thisWindow = self.deathIncrements[index: index+ self.SMAperiod]
            windowAverage = sum(thisWindow) / self.SMAperiod
#            print (thisWindow, windowAverage )
            self.SMAdata += [windowAverage]                                     # .append (windowAverage)
            index += 1

    def SMA2(self, dataList,nrPeriods):
        # returns the simple moving average of dataList over nrPeriods
        cumSum = np.cumsum(np.insert(dataList, 0, 0))                           # cumulative sum of dataList
        return (cumSum[nrPeriods:] - cumSum[:-nrPeriods]) / float(nrPeriods)    # SMA over nrPeriods

    def makeSMAarrays(self):
#        print ("======== Calculating the short and long SMAs ==========")
#        print ("\n==== len("+self.countryName +"].deathIncrements): ", 
#                                        len(self.deathIncrements))
        shortArray = self.SMA2(self.deathIncrements, self.SMAshortPeriod)       # generate the SMA short period dataset
        list2prepend = [0]* (len (self.deathIncrements) - len (shortArray))     # an SMA produces an array shorter shorter than the original dataset 
        self.SMAshortArray = np.insert(shortArray, list2prepend, 0)             # 0 arg means insert at the beginning
#        print (self.countryName+" shortSMA array len :",  len(self.SMAshortArray))            

        longArray = self.SMA2(self.deathIncrements, self.SMAlongPeriod)         # generate the SMA short period dataset
        list2prepend = [0]* (len (self.deathIncrements) - len (longArray))      # an SMA produces an array shorter shorter than the original dataset 
        self.SMAlongArray = np.insert(longArray, list2prepend, 0)               # 0 arg means insert at the beginning
#        print (self.countryName+" longSMA array len :",  len(self.SMAlongArray))            

    def nanMean (self, listy):
        #Computes the mean of a list containing float ('nan') values
        nrItems = len (listy)
        sumItems = 0
        for item in listy:
            try:
                sumItems += item                                                # cummulative sum of summable values in listy
            except:
                pass                                                            # faster than adding 0 used to avoid float ('nan' values for e.g.
#        print ("sum + len + mean", sumItems, len (listy), sumItems/len(listy))
        return sumItems/len (listy)
    # median to b calculated also.           

    def zeros2NANs (self):
        # replace all zeros with float ('nan') [Not A Number] to suppress printing of all 0 values
        nrDays = len(self.deathIncrements)
        for indx in range (1, nrDays):                                          # don't use the first one - why?
            if self.deathIncrements[indx]>(200+self.deathIncrements[indx-1]):   # attempt at filtering outliers
                pass                                                            #  not working; not used
            # print (self.deathIncrements[indx-1], self.deathIncrements[indx])  
            #self.deathIncrements[indx] = 0 # self.deathIncrements[indx-1]
        for indx in range (nrDays):                                             # leave last value to set manually
            if self.deathIncrements[indx] <= 0:                                 # zero -ve increments:resurrection currently not available :-)
               self.deathIncrements[indx] = float('nan')
        self.deathIncrements[0]=0                                               # Kludge: force a printable val so graph will start at the 0th index
        if not (self.deathIncrements [-1] > 0):
            self.deathIncrements [-1] = float('nan')                            # Kludge: force a printable val so graph will end  at the 0th index
#            self.deathIncrements += [0]                                        # Kludge: force a printable val so graph will end  at the 0th index

# ===================  </End class CountryDeaths() definition> ====================
